%--------------------------------------------------------------------------------------------------------------
% DSDLL demo: Simple soundcard input impedance meter by Stanislav Maslan
%--------------------------------------------------------------------------------------------------------------
% This simple demo measures input impedance of the soundcard channel using a reference
% resistor with known impedance. Output of the soundcard generates sinewave at selected frequencies. 
% One input channel of the soundcard is connected directly to the output to obtain reference 
% voltage U1. The second input channel is connected via reference impedance Zref to obtain voltage U2.
% The Zref together with the input impedance Zi forms a voltage dividier so the input impedance can
% be easily calculated as: Zi = Zref*U2/(U1 - U2).
% To make it a little more accurate I also implemented interchannel gain/phase correction which 
% is performed with both inputs connected together to the output channel. A correction file is 
% created and it is used to fix the gain/phase errors of the soundcard.
% The reference resistor must be chosen so it has impedance close to the input impedance. In my
% case it was 10 kOhm. If RLC meter is not available the parallel capacitance of the resitor must be
% guessed. For 0207 resistors it is usually some 0.3 pF but this is negligible error due to the
% fact the input admittance of the input channel in connected in parallel to the unknown capacitance
% of the cable. The admittance of the cable can be found by connecting another identical cable in 
% parallel to the analyzed input and reapeating the measurement. Difference bethween the mesured Cp
% should be capacitance of the cable and it should be independent to frequency.
%
% Keep in mind the L/R channels may be swapped on various soundcards so if the system shows 
% some nonsense like negative impedance try to swap inputs.
%
% The name of the DirectSound devices must be entered in the script! Type in the name as it is
% shown when the script prints out the enumerated devices. Or leave empty so it will choose your 
% system default.
%
% For optimal function set soundcard output/capture mixing rates to the same as in this script i.e. 
% prefereably 48kHz/24bit. Otherwise Windoze will resample the signals.
%
% This demo requires 'dsrec.exe', 'dsdll.dll' files in the pwd (see project WWW).
% It also requires 'ds_run.m' and 'ds_enum.m' functions in pwd.
%
%
% LICENSE
% -------
% Copyright  2016, Stanislav Maslan <s.maslan@seznam.cz>
%
% This script is free software. It comes without any warranty. You can redistribute it
% and/or modify it as you like as long as this license, the author and link to the 
% DSDLL project web page is included. 
%
% Author: Stanislav Maslan
% E-mail: s.maslan@seznam.cz
% www: http://elektronika.kvalitne.cz/SW/dsdll/dsdll_eng.html
% Revision: V1.0, 2016-08-21
% Tested: Octave 3.6.4+
%--------------------------------------------------------------------------------------------------------------
clear all;
close all;
clc;

% set pwd to this script directory otherwise it won't find the 'dsrec.exe' utility
mfld = [fileparts(mfilename('fullpath')) filesep()];
cd(mfld);

% info
printf('Simple soundcard input impedance meter by Stanislav Maslan\nusing DirectSound DLL interface DSDLL.\n\n');


% --- general setup
% corrections file name (generated by the script itself)
corr_file = 'tfcorr.csv';

% --- reference resistor impedance
% equivalent parallel resistance Rp [Ohm]
Rp_ref = 9988.7;
% equivalent parallel capacitance Cp [F]
Cp_ref = 0.25e-12;

% cable admittance to subtract from the input admittance
% for RG58 it is approximately 105 pF/m
Cp_cab = 148e-12;



% --- ADC setup
% sampling window size [s]
% use longer period for low frequencies or lower standard deviation of the measurement
T = 1.0;
% wait to settle the signals (at least few hundred ms) [s]
% this is essential to prevent transients in the waveform after playback start
% use longer values for low frequencies 
settle = 1.0;

% --- DDS setup
% sampling rate [Hz]
fs = 48000;
% desired list of frequencies to synthesize [Hz]
fx = logspace(log10(50),log10(20000),15);
% amplitude [-]
amp = 0.5;

% --- string names of the input and output DirectSound devices, leave empty to use default ones
dsinp_name = 'Radek (USB Sound Blaster HD)';
dsout_name = 'Reproduktory (USB Sound Blaster HD)';


% --- enumerate DirectSound devices, find desired devices IDs
[dsinp_id,dsout_id,dsinp,dsout] = ds_enum(dsinp_name,dsout_name);

% print list of DirectSound devices
printf('Output devices:\n');
for k = 1:numel(dsout), printf('#%d - %s\n',k - 1,dsout{k}); endfor
printf('\nCapture devices:\n');
for k = 1:numel(dsinp), printf('#%d - %s\n',k - 1,dsinp{k}); endfor
printf('\n\n');


% ====== MAIN MEASUREMENT LOOP ======
mode = 'm';
while true

  if mode == 'm'
    disp('Perform gain/phase self-calibration (press key to select)?');
    disp(' Y: yes');
    disp(' N: no, don''t perform any corrections');
    disp(' F: no, use previously made corrections from a file');
    disp(' X: exit');  
    
    while true
      a = input('','s');
      disp('');
      
      % select option
      if strcmpi(a,'y')
        mode = 'c';
        corr = 0;
        break;
      elseif strcmpi(a,'n')
        mode = 'm';
        corr = 0;      
        break;
      elseif strcmpi(a,'f')
        mode = 'm';
        corr = 1;
        break;
      elseif strcmpi(a,'x')
        return;      
      endif
    endwhile
    
  endif
  
  % display instructions
  if mode == 'c'
    printf('\nConnect L+R line-ins to L line-out, hit ENTER to continue...\n');
  else
    printf('\nConnect L line-in to L line-out directly,\n');
    printf('connect R line-in to L line-out via reference resistor,\n');
    printf('hit ENTER to continue...\n');
  endif
  kbhit();
  disp('');
  
  if corr && exist(corr_file,'file')    
    % load the corrections file
    csv = dlmread(corr_file,';');
    
    % parse it [f, real, imag]
    corr_tf = complex(csv(:,2),csv(:,3));
    corr_f = csv(:,1);
  else
    % no corrections, clear correction data
    corr_tf = [];
    corr_f = [];
    
  endif    
      

  % === SWEEP LOOP ===
  fz = [];
  Rp = [];
  Cp = [];
  for k = 1:numel(fx)
  
    % frequency of the current sweep step [Hz]
    f = fx(k);
      
    % --- calculate DDS setup so it is coherent with the loop buffer length
    % possible number of periods per the loop buffer  
    M = [1 3 5 7]; 
    % possible samples count for every M 
    N = round(fs./(f./M));
    % possible aligned (coherent) frequencies for every M
    fm = fs./(N./M);
    % select the M for lowest aligned frequency deviation from the desired one
    [v,id] = min(abs(fm - f));
    % get the aligned (coherent) frequency [Hz]
    f = fm(id);
    % get the total number of samples
    N = N(id);
    
    % calculate samples count for recorder so it is coherent with the DDS wave period
    % and so the record is at least 'T' seconds long 
    M = ceil((T*fs)/N)*N;
    
    % finally expand samples count to minimum waveform size for DSDLL,
    % becuase DS don't like too short playback loops
    if N < 16, N *= ceil(16/N); endif
    
    % just in case leave out frequencies which are above nyquist limit
    if(f > 0.45*fs)
      break;
    endif
    
    
    % --- synthesize the sine waveforms
    % generate phase vector [rad]
    p = [];
    p(:,1) = [0:N-1]*f/fs*2*pi;    
    % generate the waveform for DDS
    ugen = amp*sin(p);
    
    % --- digitize
    % send the DDS waveform to dsrec.exe and capture waveforms from a recorder
    u = ds_run(dsinp_id,dsout_id,ugen,fs,2,settle,M,1);
    
    % calculate spectra of the captured voltages, loose upper halfs of the spectra (negative frequencies)
    U = fft(u)(2:floor(M/2),:)*2/M;
    
    % DFT bin number with the analyzed frequency f
    fid = round(f/(fs/M));
    
    % extract the DFT bin with analyzed frequency f 
    Uf = U(fid,:);
      
    if mode == 'c'
      % --- calibration mode
      
      % complex channel to channel transfer
      tfx = Uf(1)/Uf(2);      
            
      % store R to L complex transfer error
      corr_tf(k,1) = tfx;
      % store frequency
      corr_f(k,1) = f;
      
      printf('f = %7.1f Hz, gain = %8.6f , phase = %8.6f rad\n',f,abs(tfx),arg(tfx));
      
    else
      % --- measurement mode
    
      if corr
        % correction enabled - perform        
        if min(corr_f) <= f && max(corr_f) >= f
          
          % valid correction data for the frequency, interpolate in the correction table          
          tfx = interp1(corr_f,corr_tf,f,'spline');
          
          % apply correction
          Uf(2) .*= tfx;          
          
        endif
        
      endif
    
      % build reference resistor complex impedance
      Zref = 1./(1/Rp_ref + j*Cp_ref*2*pi*f);
      
      % calculate complex input impedance 
      Zx = Zref*Uf(2)/(Uf(1) - Uf(2));          
      
      % convert the input impedance to Rp-Cp equivalent circuit in [kOhm-pF]
      Rp(k,1) = 0.001./real(1/Zx);
      Cp(k,1) = 1e12*(imag(1/Zx)/2/pi/f - Cp_cab);
      fz(k,1) = f;
      
      % display results
      printf('f = %7.1f Hz, Rp = %10.4f kOhm, Cp = %8.2f pF\n',fz(k),Rp(k),Cp(k));
    endif    
    
    
    if numel(kbhit(1))
      % leave if key pressed...    
      break;
    endif 
  
  endfor
  
  if mode == 'c'
    % --- calibration mode selected, store the newly measured corrections to a file
    csv = [corr_f real(corr_tf) imag(corr_tf)];
    dlmwrite(corr_file,csv,';');
    
    % switch to measurement mode
    mode = 'm';
    
    disp('');
             
  elseif mode == 'm'
    % --- measuremet mode - leave
    break;
  endif
  
endwhile

% --- plot the results
subplot(2,1,1);
semilogx(fx,Rp,'LineWidth',1.5);
%xlabel('f [Hz]');
ylabel('Rp [kOhm]');
title('Input admittance of the Creative X-FI HD USB');
grid on;

subplot(2,1,2);
loglog(fx,Cp,'LineWidth',1.5);
xlabel('f [Hz]');
ylabel('Cp [pF]');
grid on;


%csv = dlmread('z_e4980a.csv',';');
%ef = csv(:,1);
%eCp = csv(:,2);
%eRp = csv(:,3);
%subplot(2,1,1);
%hold on;
%plot(ef,eRp,'ro','LineWidth',2,'MarkerSize',4);
%hold off;
%subplot(2,1,2);
%hold on;
%plot(ef,eCp,'ro','LineWidth',2,'MarkerSize',4);
%hold off;
%legend('measured by dsdll meter','measured by RLC meter E4980A');

%print('inpz.png','-S640,400','-F:7','-dpng');